"""
Infra for JSON Path-like expressions and finding items in data hiearchy.
"""

# pylint: disable=protected-access
from __future__ import annotations

import dataclasses as dc
import re
from abc import ABC
from collections import namedtuple
from typing import Any, Callable, Iterator


@dc.dataclass(frozen=True, kw_only=True)
class _Accessor(ABC):
    """
    Base class for building JSON Path-like expression.
    """

    parent: _Accessor | None
    container_type: type

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        """
        Args:
            container: the data structure to access into
            yield_path: whether to yield a singular path to the item (or items) that are being accessed
            lenient: whether to allow out-of-bounds indexing or missing dict key references to raise
                IndexError or KeyError, respectively
        Returns:
            An iterator to tuples, where there the first field of the tuple is the item being accessed
            at this point, and the second field of the tuple is either ``None``, or a function taking
            a single ``parent`` argument, to create paths leading to those items being accessed.
        """
        raise NotImplementedError("Root accessor should not be invoked")

    def _is_singular(self) -> bool:
        """
        Returns: if this accessor (excluding any parent from consideration) can only
            ever return a single item.
        """
        return True

    def is_singular(self) -> bool:
        """
        Returns: if this path (this accessor and its parents, transitively) can only
            ever return a single item. E.g. a slice expression in a path makes it non-singular,
            even if the start/stop combination otherwise would say it's singular, like ``[1:2]``.
        """
        return self._is_singular() and (
            self.parent is None or self.parent.is_singular()
        )

    def _representation(self) -> str:
        """
        Represention for this accessor (excluding any parent)
        """
        raise NotImplementedError("Root accessor should not be invoked")

    def _check_container_type(self, container: list | dict):
        """
        Performs a check that a container to access is of the type mandated by
        the declared container_type of an accessor instance.
        """
        if not isinstance(  # pylint: disable=isinstance-second-argument-not-valid-type
            container, self.container_type
        ):
            raise ValueError(
                f"Expected {self.container_type}, got {type(container)} at {self}"
            )

    def _accessors(self) -> list[_Accessor]:
        """
        List of accessors, from root to this, recursively collected.
        """
        if self.parent is None:
            # Symmetry would suggest here to `return [self]`, however,
            # we cheat here: this instance shall be the root (=first accessor
            # to traverse by), not the parent
            return []
        return self.parent._accessors() + [self]

    def __str__(self):
        accessors = self._accessors()
        return "$" + "".join(a._representation() for a in accessors)

    def __getattr__(self, specification):
        """
        JSON Path-like expression builder infra.

        Overloaded for dict key access.
        """
        return _DictKeyAccessor(parent=self, key=specification)

    def __getitem__(self, specification):
        """
        JSON Path-like expression builder infra.

        Overloaded for list (index, slice) and various dict key access.
        """
        if isinstance(specification, str):
            if specification == "*":
                return _ListSliceAccessor(parent=self, slice_=slice(None, None, None))
            if specification.isidentifier():
                return _DictKeyAccessor(parent=self, key=specification)
            key_regex = re.compile(specification)
            return _DictKeyRegexAccessor(parent=self, key_regex=key_regex)
        if isinstance(specification, int):
            return _ListIndexAccessor(parent=self, index=specification)
        if isinstance(specification, slice):
            return _ListSliceAccessor(parent=self, slice_=specification)
        raise ValueError(f"Unsupported indexing expression {specification}")

    def __call__(self, *paths, **named_paths):
        """
        JSON Path-like expression builder infra.

        Overloaded for sub-item selection (vanilla or named tuple).
        """
        if len(paths) > 0 and len(named_paths) > 0:
            raise NotImplementedError(
                "Either all sub-selections are to be named, or none of them."
            )

        def verify_paths(paths):
            not_accessors = [path for path in paths if not isinstance(path, _Accessor)]
            if len(not_accessors) > 0:
                raise ValueError(f"Not paths: {not_accessors}")
            not_singulars = [path for path in paths if not path.is_singular()]
            if len(not_singulars) > 0:
                raise ValueError(
                    f"Not singular paths (could point to multiple items): {not_singulars}"
                )

        if len(paths) > 0:
            verify_paths(paths)
            return _SubHiearchyAccessor(
                parent=self, sub_accessors=paths, tuple_ctor=tuple
            )
        if len(named_paths) > 0:
            paths = list(named_paths.values())
            verify_paths(paths)
            named_tuple = namedtuple("SubSelect", list(named_paths.keys()))
            return _SubHiearchyAccessor(
                parent=self,
                sub_accessors=paths,
                tuple_ctor=lambda values: named_tuple(*values),
            )
        raise NotImplementedError("Sub-selection cannot be empty")


@dc.dataclass(frozen=True, kw_only=True)
class _DictKeyAccessor(_Accessor):
    """Accesses a value of a dict container by a key"""

    key: str
    container_type: type = dict

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        if lenient and self.key not in container:
            yield from iter([])
        elif yield_path:
            yield container[self.key], (lambda parent: _DictKeyAccessor(parent=parent, key=self.key))  # type: ignore
        else:
            yield container[self.key], None  # type: ignore

    def _representation(self) -> str:
        return f".{self.key}"


@dc.dataclass(frozen=True, kw_only=True)
class _DictKeyRegexAccessor(_Accessor):
    """Accesses a value or values of a dict container by a regex matching keys"""

    key_regex: re.Pattern
    container_type: type = dict

    def _is_singular(self) -> bool:
        return False

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        for key, value in container.items():  # type: ignore
            if self.key_regex.findall(key):  # pylint: disable=no-member
                if yield_path:
                    yield value, lambda parent: _DictKeyAccessor(
                        parent=parent, key=key  # pylint: disable=cell-var-from-loop
                    )
                else:
                    yield value, None

    def _representation(self) -> str:
        return f"[{self.key_regex.pattern}]"  # pylint: disable=no-member


@dc.dataclass(frozen=True, kw_only=True)
class _ListIndexAccessor(_Accessor):
    """Accesses an item of a list by an index"""

    index: int
    container_type: type = list

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        self._check_container_type(container)
        if lenient and (self.index >= len(container) or self.index < -len(container)):
            yield from iter([])
        elif yield_path:
            if self.index >= 0:
                yield container[self.index], lambda parent: _ListIndexAccessor(
                    parent=parent, index=self.index
                )
            else:
                yield container[self.index], lambda parent: _ListIndexAccessor(
                    parent=parent, index=len(container) + self.index
                )
        else:
            yield container[self.index], None

    def _representation(self) -> str:
        return f"[{self.index}]"


@dc.dataclass(frozen=True, kw_only=True)
class _ListSliceAccessor(_Accessor):
    """Accesses items of a list by a slice"""

    slice_: slice
    container_type: type = list

    def _is_singular(self) -> bool:
        return False

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        _ = lenient  # unused
        self._check_container_type(container)
        if yield_path:

            def create_accessor_ctor(index: int) -> _Accessor:
                return lambda parent: _ListIndexAccessor(parent=parent, index=index)  # type: ignore

            slice_indices = self.slice_.indices(len(container))

            yield from (
                (item, create_accessor_ctor(current_index))
                for current_index, item in zip(
                    range(slice_indices[0], slice_indices[1], slice_indices[2]),
                    container[self.slice_],
                )
            )
        else:
            yield from ((item, None) for item in container[self.slice_])

    def _representation(self) -> str:
        repr_ = (
            (f"[{self.slice_.start}" if self.slice_.start is not None else "[")
            + (f":{self.slice_.stop}" if self.slice_.stop is not None else ":")
            + (f":{self.slice_.step}]" if self.slice_.step is not None else "]")
        )
        return repr_


@dc.dataclass(frozen=True, kw_only=True)
class _SubHiearchyAccessor(_Accessor):
    """
    Instead of returning an entire item (of a list), it constructs a tuple based on sub-JSON Path-like expressions.
    """

    sub_accessors: list[_Accessor]
    tuple_ctor: Any
    container_type: type = dict

    def _access(
        self, container: list | dict, *, yield_path: bool = False, lenient: bool = False
    ) -> Iterator[tuple[Any, Callable[[_Accessor], _Accessor] | None]]:
        _ = lenient  # unused
        self._check_container_type(container)
        items = [
            find_next(container, sub_accessor, default=None)
            for sub_accessor in self.sub_accessors
        ]
        if yield_path:
            yield self.tuple_ctor(items), lambda parent: _SubHiearchyAccessor(
                parent=parent,
                sub_accessors=self.sub_accessors,
                tuple_ctor=self.tuple_ctor,
            )
        else:
            yield self.tuple_ctor(items), None

    def _representation(self):
        return (
            f"({', '.join(str(sub_accessor) for sub_accessor in self.sub_accessors)})"
        )


JP = _Accessor(parent=None, container_type=type(None))


def find_all(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Finds all matching items in a JSON-like data hierarchy (lists of / dicts of / values) based
    on a JSON Path-like specification. Technically, it iterates over the matching items.

    Args:
        path: JSON Path-like expression, specifying what item (or items) to match & iterate over
        with_path: whether to yield accurate, JSON Path-like pointer objects to items found
        lenient: whether to allow out of bound indices or missing keys, or raise ``IndexError`` and
            ``KeyError`` exceptions, respectivately.
    """
    all_accessors = list(path._accessors())
    stack = [(root_data, all_accessors, JP if with_path else None)]

    while len(stack) > 0:
        data, accessors, current_accessor = stack.pop()
        if len(accessors) == 0:  # leaf item
            if with_path:
                yield current_accessor, data
            else:
                yield data
        else:
            accessor = accessors[0]

            # With a stack content [...] and items A, B, C iterated by accessor._access(...)
            # we want the following stack content: [..., C*, B*, A*]
            # - where A*, B*, C* are tuples created for A, B, C respectively
            # the point is that we want to process, in the next loop, in the order A*. B*, C*.
            #
            # We don't want to do an equivalent `for ... in reversed(list(accessor._access(...))):`,
            # as reversal requires constructing a full list first in order to reverse the order.
            #
            # Inserting into the Nth position (N is current length of stack) achieves the same.
            stack_insert_position = len(stack)
            sub_tuples = accessor._access(data, yield_path=with_path, lenient=lenient)  # type: ignore
            for sub_data, accessor_ctor in sub_tuples:
                new_accessor = accessor_ctor(current_accessor) if accessor_ctor is not None else None  # type: ignore
                stack.insert(
                    stack_insert_position, (sub_data, accessors[1:], new_accessor)
                )


def find_next(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    **kwargs,
):
    """
    Shorthand for ``next(find_all(...))``. Also takes a keyword argument, ``default``,
    to delegate it to the ``next(..., default=...)`` call, if defined.
    """
    if "default" in kwargs:
        default = kwargs["default"]
        try:
            # we don't want to pass the default value as `next(..., default)` ...
            return next(find_all(root_data, path, with_path=with_path, lenient=True))
        except StopIteration:
            # ... because we need to return None for path, if with_path=True
            return (None, default) if with_path else default
    else:
        return next(find_all(root_data, path, with_path=with_path))


def _find_all_with_multiplicity(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Workhorse utility function for finding both unique and duplicate items.
    """
    item_func = (lambda i: i[1]) if with_path else (lambda i: i)

    known_items: dict = {}

    for tup in find_all(root_data, path, with_path=with_path, lenient=lenient):
        item = item_func(tup)
        if item in known_items:
            known_items[item] += 1
        else:
            known_items[item] = 0
        yield tup, known_items[item]


def find_all_unique(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Yields unique elemnents, with or without paths.
    """
    yield from (
        item
        for item, multiplicity in _find_all_with_multiplicity(
            root_data, path, with_path=with_path, lenient=lenient
        )
        if multiplicity == 0
    )


def find_all_duplicate(
    root_data: list | dict | str | int | float | bool,
    path: _Accessor,
    *,
    with_path: bool = False,
    lenient: bool = False,
):
    """
    Yields duplicate elemnents, with or without paths.
    """
    yield from (
        item
        for item, multiplicity in _find_all_with_multiplicity(
            root_data, path, with_path=with_path, lenient=lenient
        )
        if multiplicity > 0
    )



import pytest

from aweson import JP, find_all, find_all_duplicate, find_all_unique, find_next


@pytest.mark.parametrize(
    "jp,stringfied",
    [
        (JP, "$"),
        (JP.hello, "$.hello"),
        (JP.hello.world, "$.hello.world"),
        (JP["hello"], "$.hello"),
        (JP["hello"]["world"], "$.hello.world"),
        (JP[0], "$[0]"),
        (JP[42], "$[42]"),
        (JP[-1], "$[-1]"),
        (JP[5][42], "$[5][42]"),
        (JP[:], "$[:]"),
        (JP["*"], "$[:]"),
        (JP[::], "$[:]"),
        (JP[1:], "$[1:]"),
        (JP[1::], "$[1:]"),
        (JP[:2], "$[:2]"),
        (JP[:2:], "$[:2]"),
        (JP[::-1], "$[::-1]"),
        (JP[1:12:-1], "$[1:12:-1]"),
        (JP[5].hello, "$[5].hello"),
        (JP.hello[5], "$.hello[5]"),
        (JP["hello|hi"].id, "$[hello|hi].id"),
        (JP[5].hello[42].world, "$[5].hello[42].world"),
        (JP.hello[5].world[42], "$.hello[5].world[42]"),
        (JP["hello"][5]["world"][42], "$.hello[5].world[42]"),
        (JP.hello(JP.world, JP.hi[0]), "$.hello($.world, $.hi[0])"),
    ],
)
def test_jp_stringification(jp, stringfied):
    assert str(jp) == stringfied


@pytest.mark.parametrize(
    "jp, is_singular",
    [
        (JP, True),
        (JP.hello, True),
        (JP.hello[0], True),
        (JP.hello[-1], True),
        (JP["hello"], True),
        (JP[:], False),
        (JP["*"], False),
        (JP[1:2], False),
        (JP.hello[:], False),
        (JP[1:2].hello, False),
    ],
)
def test_path_is_singular(jp, is_singular):
    assert jp.is_singular() == is_singular


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ("string", JP, ["string"]),
        (5, JP, [5]),
        (5.13, JP, [5.13]),
        (True, JP, [True]),
        ({"hello": 42}, JP, [{"hello": 42}]),
        ([5, 42], JP, [[5, 42]]),
    ],
)
def test_find_all_for_root_path(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ({"hello": 42, "hi": "irrelevant"}, JP.hello, [42]),
        ({"hello": {"world": 42}, "hi": "irrelevant"}, JP.hello.world, [42]),
        ({"hello": 42, "hi": "irrelevant"}, JP["hello"], [42]),
        ({"hello": {"world": 42}, "hi": "irrelevant"}, JP["hello"]["world"], [42]),
    ],
)
def test_find_all_simple_key_based_dict_access(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ([5, 42], JP[0], [5]),
        ([5, 42], JP[-1], [42]),
        ([5, 42, 137], JP[:], [5, 42, 137]),
        ([5, 42, 137], JP["*"], [5, 42, 137]),
        ([5, 42, 137], JP[1:], [42, 137]),
        ([5, 42, 137], JP[:1], [5]),
        ([5, 42, 137], JP[::-1], [137, 42, 5]),
    ],
)
def test_find_all_list_access_by_index_and_slice(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP[".*"][:].name,
            ["red delicious", "punakaneli", "wilhelm", "conference"],
        ),
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP["apple|pear"][:].name,
            ["red delicious", "punakaneli", "wilhelm", "conference"],
        ),
    ],
)
def test_find_all_access_by_key_regex(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            {
                "apple": [{"name": "red delicious"}, {"name": "punakaneli"}],
                "pear": [{"name": "wilhelm"}, {"name": "conference"}],
            },
            JP[".*"][:].name,
            [
                (JP.apple[0].name, "red delicious"),
                (JP.apple[1].name, "punakaneli"),
                (JP.pear[0].name, "wilhelm"),
                (JP.pear[1].name, "conference"),
            ],
        ),
    ],
)
def test_find_all_access_by_key_regex_with_path(content, jp, expected_items):
    items = list(find_all(content, jp, with_path=True))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        ({"hello": [5, 42]}, JP.hello[1], [42]),
        ([{"hello": 42}, {"hello": 5}], JP[1].hello, [5]),
        ([{"hello": 42}, {"hello": 5}], JP[1]["hello"], [5]),
        (
            {"l1": [{"detail": {"data": 5}}, {"detail": {"data": 42}}]},
            JP.l1[:].detail.data,
            [5, 42],
        ),
    ],
)
def test_find_all_mixed_hierarchy(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](JP.hello, JP.hi[1]),
            [("world", 42), ("mundo", -42)],
        ),
        (
            {
                "employees": [
                    {"name": "Doe, John", "age": 32, "account": "johndoe"},
                    {"name": "Doe, Jane", "age": -23, "account": "janedoe"},
                    {"name": "Deer, Jude", "age": 42, "account": "judedeer"},
                ]
            },
            JP.employees[:2](JP.account, JP.name),
            [("johndoe", "Doe, John"), ("janedoe", "Doe, Jane")],
        ),
    ],
)
def test_find_all_sub_selection(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,expected_items",
    [
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](id_=JP.hello, detail=JP.hi[1]),
            [("world", 42), ("mundo", -42)],
        ),
        (
            {
                "employees": [
                    {"name": "Doe, John", "age": 32, "account": "johndoe"},
                    {"name": "Doe, Jane", "age": -23, "account": "janedoe"},
                    {"name": "Deer, Jude", "age": 42, "account": "judedeer"},
                ]
            },
            JP.employees[:2](id_=JP.account, detail=JP.name),
            [("johndoe", "Doe, John"), ("janedoe", "Doe, Jane")],
        ),
    ],
)
def test_find_all_sub_selection_named_tuple(content, jp, expected_items):
    items = list(find_all(content, jp))
    assert len(items) == len(expected_items)
    for item, expected_item in zip(items, expected_items):
        assert item.id_ == expected_item[0]
        assert item.detail == expected_item[1]


def test_find_all_demo_transformation():
    list_of_entities = [
        {"id": 5, "value": "five"},
        {"id": 42, "value": "life, universe and everything"},
        {"id": 137, "value": "137"},
    ]

    id_to_value_map = {
        id_: value for id_, value in find_all(list_of_entities, JP[:](JP.id, JP.value))
    }
    assert id_to_value_map == {
        5: "five",
        42: "life, universe and everything",
        137: "137",
    }


def test_find_all_demo_transformation_named_tuples():
    list_of_entities = [
        {"id": 5, "value": "five"},
        {"id": 42, "value": "life, universe and everything"},
        {"id": 137, "value": "137"},
    ]

    id_to_value_map = {
        tup.id_: tup.value
        for tup in find_all(list_of_entities, JP[:](id_=JP.id, value=JP.value))
    }
    assert id_to_value_map == {
        5: "five",
        42: "life, universe and everything",
        137: "137",
    }


def test_sub_selection_either_all_unnamed_or_all_named_fields():
    with pytest.raises(NotImplementedError):
        JP.hello(JP.some_field, named=JP.other_field)


def test_sub_selection_must_not_be_empty():
    with pytest.raises(NotImplementedError):
        JP.hello()


def test_sub_selection_must_be_path():
    with pytest.raises(ValueError):
        JP[:](JP.id, "somestring")

    with pytest.raises(ValueError):
        JP[:](id_=JP.id, detail="somestring")


def test_sub_selection_must_be_singular():
    with pytest.raises(ValueError):
        JP[:](JP.id, JP.detail[:])

    with pytest.raises(ValueError):
        JP[:](id_=JP.id, detail=JP.detail[:])


@pytest.mark.parametrize(
    "content,jp,expected_paths_and_items",
    [
        ("string", JP, [(JP, "string")]),
        (
            {"hello": {"world": 42}, "hi": "irrelevant"},
            JP.hello.world,
            [(JP.hello.world, 42)],
        ),
        ([5, 42], JP[-1], [(JP[1], 42)]),
        ([5, 42, 137], JP[1:], [(JP[1], 42), (JP[2], 137)]),
        ([5, 42, 137], JP[::-1], [(JP[2], 137), (JP[1], 42), (JP[0], 5)]),
        ([{"hello": 42}, {"hello": 5}], JP[1:]["hello"], [(JP[1].hello, 5)]),
        (
            [{"hello": 42}, {"hello": 5}],
            JP[-1::-1]["hello"],
            [(JP[1].hello, 5), (JP[0].hello, 42)],
        ),
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[:](JP.hello, JP.hi[1]),
            [
                (JP[0](JP.hello, JP.hi[1]), ("world", 42)),
                (JP[1](JP.hello, JP.hi[1]), ("mundo", -42)),
            ],
        ),
        (
            [
                {"hello": "world", "hi": [5, 42, 137]},
                {"hello": "mundo", "hi": [-5, -42, -137]},
            ],
            JP[1:0:-1](JP.hello, JP.hi[1]),
            [(JP[1](JP.hello, JP.hi[1]), ("mundo", -42))],
        ),
    ],
)
def test_find_all_with_paths(content, jp, expected_paths_and_items):
    paths_and_items = list(find_all(content, jp, with_path=True))
    assert len(paths_and_items) == len(expected_paths_and_items)
    for path_and_item, expected_path_and_item in zip(
        paths_and_items, expected_paths_and_items
    ):

        path, item = path_and_item

        # The path yielded alongside the item points to the item itself ...
        assert item == next(find_all(content, path))

        # ... but let's check it against explicit expectations, too.
        expected_path, expected_item = expected_path_and_item
        assert item == expected_item
        assert str(path) == str(expected_path)


@pytest.mark.parametrize(
    "content, jp",
    [
        ([], JP[0]),
        ([5, 42], JP[2]),
        ({"hello": [5, 42]}, JP.hello[2]),
        ({"hello": [5, 42]}, JP["hello"][2]),
    ],
)
def test_find_all_index_error(content, jp):
    with pytest.raises(IndexError):
        list(find_all(content, jp))


@pytest.mark.parametrize(
    "content, jp",
    [
        ([], JP[0]),
        ([5, 42], JP[2]),
        ([5, 42], JP[-3]),
        ({"hello": [5, 42]}, JP.hello[2]),
        ({"hello": [5, 42]}, JP["hello"][2]),
    ],
)
def test_lenient_find_all_yields_nothing_on_nonexistent_indexes(content, jp):
    assert list(find_all(content, jp, lenient=True)) == []


@pytest.mark.parametrize(
    "content, jp",
    [
        ({}, JP.hello),
        ({"hello": {"world": 42}}, JP.hello.hi),
        ({"hello": {"world": 42}}, JP["hello"]["hi"]),
    ],
)
def test_find_all_key_error(content, jp):
    with pytest.raises(KeyError):
        list(find_all(content, jp))


@pytest.mark.parametrize(
    "content, jp",
    [
        ({}, JP.hello),
        ({"hello": {"world": 42}}, JP.hello.hi),
        ({"hello": {"world": 42}}, JP["hello"]["hi"]),
    ],
)
def test_lenient_find_all_yields_nothing_on_nonexistent_keys(content, jp):
    assert list(find_all(content, jp, lenient=True)) == []


@pytest.mark.parametrize(
    "content,jp,with_path,expected_value",
    [
        (
            [{"hello": 5}, {"hello": 42}],
            JP[:].hello,
            False,
            5,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[1].hello,
            False,
            42,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[:].hello,
            True,
            (JP[0].hello, 5),
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[1].hello,
            True,
            (JP[1].hello, 42),
        ),
    ],
)
def test_find_next(content, jp, with_path, expected_value):
    assert find_next(content, jp, with_path=with_path) == expected_value


@pytest.mark.parametrize(
    "content,jp,with_path,default",
    [
        # cases: slice with empty list
        ([], JP[:].hello, False, 17),
        ([], JP[:].hello, True, (None, 17)),
        # cases: ignoring index error
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hello,
            False,
            5,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hello,
            True,
            (None, 5),
        ),
        # cases: ignoring key error
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hi,
            False,
            5,
        ),
        (
            [{"hello": 5}, {"hello": 42}],
            JP[2].hi,
            True,
            (None, 5),
        ),
    ],
)
def test_find_next_with_default(content, jp, with_path, default):
    assert find_next(content, jp, with_path=with_path, default=default)


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        ([1, 2, 1, 3, -22, 3], JP[:], False, [1, 2, 3, -22]),
        (
            [1, 2, 1, 3, -22, 3],
            JP[:],
            True,
            [(JP[0], 1), (JP[1], 2), (JP[3], 3), (JP[4], -22)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            False,
            [1, 2, 3, -22],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            True,
            [(JP[0].hello, 1), (JP[1].hello, 2), (JP[3].hello, 3), (JP[4].hello, -22)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:4].hello,
            False,
            [1, 2, 3],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:4].hello,
            True,
            [(JP[0].hello, 1), (JP[1].hello, 2), (JP[3].hello, 3)],
        ),
    ],
)
def test_find_all_unique(content, jp, with_path, expected_items):
    items = list(find_all_unique(content, jp, with_path=with_path))
    assert items == expected_items


@pytest.mark.parametrize(
    "content,jp,with_path,expected_items",
    [
        ([1, 2, 1, 3, -22, 3], JP[:], False, [1, 3]),
        ([1, 2, 1, 3, -22, 3], JP[:], True, [(JP[2], 1), (JP[5], 3)]),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            False,
            [1, 3],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:].hello,
            True,
            [(JP[2].hello, 1), (JP[5].hello, 3)],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:3].hello,
            False,
            [1],
        ),
        (
            [
                {"hello": 1},
                {"hello": 2},
                {"hello": 1},
                {"hello": 3},
                {"hello": -22},
                {"hello": 3},
            ],
            JP[:3].hello,
            True,
            [(JP[2].hello, 1)],
        ),
    ],
)
def test_find_all_duplicates(content, jp, with_path, expected_items):
    items = list(find_all_duplicate(content, jp, with_path=with_path))
    assert items == expected_items
